# -*- coding: utf-8 -*-
#
#------------------------------------------------------------------------------
# TrasMiSIG  -  Software libre para consultar y visualizar rutas ptimas en una
# red de transporte masivo compuesta por estaciones y segmentos. Utiliza
# PostgreSQL/PostGIS y PyQGIS.
#
# Original sources Copyright (c) 2006 by Tim Sutton
# Ported to Python by Martin Dobias
# Copyright (C) 2009  Germn Carrillo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#------------------------------------------------------------------------------

import sys
import operator # Para hacer un sort

from PyQt4.QtCore import *
from PyQt4.QtGui import *

from qgis.core import *
from qgis.gui import *

from config import *
from trasmisig_ui import Ui_TrasMiSIG
from conexion_postgresql import ConexionPgSQL

class Visor( QMainWindow, Ui_TrasMiSIG ):
    """ Clase Visor """
    def __init__( self ):
        """ Inicializador de la clase Visor """

        # Inicializar la app de Qt4
        QMainWindow.__init__( self )
        self.setupUi( self )

        # Splash Screen
        splash = QSplashScreen( QPixmap( "recursos/imagenes/splash.png" ) )
        splash.show()
        splash.showMessage( "Cargando interfaz...",
            Qt.AlignBottom | Qt.AlignHCenter )

        # Crear el Map Canvas
        self.canvas = QgsMapCanvas()
        self.canvas.setCanvasColor( QColor( 204,255,204 ) )
        self.canvas.enableAntiAliasing( True )
        self.canvas.useImageToRender(False)
        self.canvas.show()

        # Se agrega el Map canvas a la ventana principal, dentro del marco
        self.layout = QVBoxLayout( self.frameMapa )
        self.layout.addWidget( self.canvas )

        # Crear las herramientas (tools) para el mapa
        self.toolAcercar = QgsMapToolZoom( self.canvas, False ) #false = Acercar
        self.toolAlejar = QgsMapToolZoom( self.canvas, True ) #true = Alejar
        self.toolPanear = QgsMapToolPan( self.canvas )
        self.toolIdentificar = MapTool( self.canvas, "Identificar" )
        self.toolSeleccionar = MapTool( self.canvas, "Seleccionar" )

        # Variables iniciales
        self.bSelOrigen = False # Flag para elegir estacin de origen o destino
        self.sOrigen = ""
        self.sDestino = ""

        # SIGNALS - SLOTS
        self.connect( self.btnCalcularRutas, SIGNAL("clicked()"),
            self.calcularRutas )
        self.connect( self.btnAcercar, SIGNAL( "clicked()" ), self.acercar )
        self.connect( self.btnAlejar, SIGNAL( "clicked()" ), self.alejar )
        self.connect( self.btnPanear, SIGNAL( "clicked()" ), self.panear )
        self.connect( self.btnVistaCompleta, SIGNAL( "clicked()" ),
            self.vistaCompleta )
        self.connect( self.btnIdentificar, SIGNAL( "clicked()" ),
            self.identificar )
        self.connect( self.btnSeleccionar, SIGNAL( "clicked()" ),
            self.seleccionar )
        self.connect( self.btnGuardarImagen, SIGNAL( "clicked()" ),
            self.guardarImagen )
        self.connect( self.btnNuevaRuta, SIGNAL( "clicked()" ),
            self.nuevaRuta )

        self.connect( self.canvas, SIGNAL( "xyCoordinates(QgsPoint&)" ),
            self.mostrarMapTip )
        self.connect( self.toolIdentificar, SIGNAL( "identificarEstacion" ),
            self.identificarEstacion )
        self.connect( self.toolSeleccionar, SIGNAL( "seleccionarEstacion" ),
            self.seleccionarEstacion )

        QObject.connect( self.cboOrigen,
            SIGNAL( "currentIndexChanged(QString)" ), self.onChangecboOrigen )
        QObject.connect( self.cboDestino,
            SIGNAL( "currentIndexChanged(QString)" ), self.onChangecboDestino )

        # Ajustes interfaz
        self.setWindowTitle( "TrasMiSIG v." + VERSION )
        self.lblAcercaDeTitulo.setText( "TrasMiSIG v." + VERSION )
        self.tblResultados.setCornerButtonEnabled( False )
        self.tblResultados.verticalHeader().setClickable( False )
        self.grbNuevaRuta.setVisible( False )
        self.grbMensajes.setVisible( False )
        self.grbIdentificar.setVisible( False )
        self.grbResultados.setVisible( False )
        self.setTabOrder( self.cboOrigen, self.cboDestino )
        self.setTabOrder( self.cboDestino, self.cboDia )
        self.setTabOrder( self.cboDia, self.cboHora )
        self.setTabOrder( self.cboHora, self.cboMinutos )
        self.setTabOrder( self.cboMinutos, self.btnCalcularRutas )
        self.cargarLeyenda()
        self.cargarNorte()
        self.cargarProgressBar()

        self.layers = [] # Lista de capas del canvas
        self.bCapasRutaCreadas = False # Flag para saber si se deben crear o
                                       # actualizar las capas de la ruta
        self.utils = Utilidades( self )

        # Cargar capas y datos e informarlo en el Splash Screen
        splash.showMessage( "Cargando capas...",
            Qt.AlignBottom | Qt.AlignHCenter )
        self.cargarCapas()
        self.utils.mostrarLabels( self.utils.obtenerCapa( self.canvas,
            "EtiquetasPortalIzq" ), "nombreestacion" )
        self.utils.mostrarLabels( self.utils.obtenerCapa( self.canvas,
            "EtiquetasPortalDer" ), "nombreestacion" )

        self.capaEstacion = self.utils.obtenerCapa( self.canvas, "Estacion" )

        # Configuracin del MapTip para mostrar el nombre de la estacin
        self.capaEstacion.setDisplayField( "nombreestacion" )
        self.mapTip = QgsMapTip()
        self.mapToPixel = self.canvas.getCoordinateTransform()

        splash.showMessage( "Cargando datos iniciales...",
            Qt.AlignBottom | Qt.AlignHCenter )
        self.conexionBD = ConexionPgSQL()
        self.cargarCombosOrigenDestino()

        self.canvas.setMapTool( self.toolAcercar ) # Tool activa por defecto
        splash.finish(self)

    @pyqtSignature("")
    def acercar( self ):
        """ Acercar """
        self.canvas.setMapTool( self.toolAcercar )

    @pyqtSignature("")
    def alejar( self ):
        """ Alejar """
        self.canvas.setMapTool( self.toolAlejar )

    @pyqtSignature("")
    def panear( self ):
        """ Panear """
        self.canvas.setMapTool( self.toolPanear )

    @pyqtSignature("")
    def identificar( self ):
        """ Identificar Estacin por medio de click en el mapa """
        self.canvas.setMapTool( self.toolIdentificar )

    @pyqtSignature("")
    def seleccionar( self ):
        """ Seleccionar Estacin por medio de click en el mapa """
        self.canvas.setMapTool( self.toolSeleccionar )
        self.grbMensajes.setVisible( True )
        self.grbAcercaDe.setVisible( False )
        self.grbResultados.setVisible( False )
        self.grbIdentificar.setVisible( False )
        self.mostrarMensajeInferior( 1 )

    @pyqtSignature("")
    def vistaCompleta( self ):
        """ Vista total """
        self.canvas.zoomToFullExtent()

    @pyqtSignature("")
    def guardarImagen( self ):
        """ Guardar el mapa como imagen (png, jpg o bmp)"""
        bRuta = 0
        while ( not bRuta ): # Se valida la ruta ingresada
            rutaImagen = QFileDialog.getSaveFileName( self,
                "Guardar mapa como imagen.",
                self.utils.ultimaRutaExportarMapaAImagen,
                "Imgenes (*.png *.jpg *.bmp)" )

            if rutaImagen:
                imagenInfo = QFileInfo( rutaImagen )
                self.utils.ultimaRutaExportarMapaAImagen = imagenInfo.absolutePath()

                # Si el archivo no tiene extensin
                if ( rutaImagen[ -4 ] != "." or rutaImagen[ -5 ] == "/" ):
                    bRuta = 0
                    self.utils.mostrarMensaje( 'Exportar mapa a imagen',
                        'Debes escribir el nombre de la imagen y su extensin.',
                        QMessageBox.Warning )
                else:
                    # Si la extensin no coincide
                    if ( rutaImagen[ -3: ] != "png" \
                        and rutaImagen[ -3: ] != "jpg" \
                        and rutaImagen[ -3: ] != "bmp" ):
                        bRuta = 0
                        self.utils.mostrarMensaje( 'Exportar mapa a imagen',
                            'La extensin que escribiste no es correcta. ' \
                            + '\nDebes corregirla.', QMessageBox.Warning )
                    else: # Exportar
                        bRuta = 1
                        self.canvas.saveAsImage( rutaImagen,
                            self.canvas.canvasPixmap(),
                            str( rutaImagen[ -3: ] ).upper() )
                        self.utils.mostrarMensaje( 'Exportar mapa a imagen',
                            'El mapa ha sido exportado exitosamente.',
                            QMessageBox.Information )
            else: # Click en cancelar
                bRuta = 1

    def cargarCapas( self ):
        """ Agregar capas iniciales de PostGIS por nombre """
        self.cargarCapaPostGIS( "bogota" )
        self.cargarCapaPostGIS( "segmento" )
        self.cargarCapaPostGIS( "estacion" )

        # Para etiquetas iniciales
        self.cargarCapaPostGIS( "estacion", "EtiquetasPortalIzq",
            "tipoestacion = 'Portal' AND nombreestacion <> 'Portal del Norte'" \
            + " AND nombreestacion <> 'Las Aguas'" )
        self.cargarCapaPostGIS( "estacion", "EtiquetasPortalDer",
            "nombreestacion = 'Las Aguas' OR nombreestacion = " \
            + "'Portal del Norte'" )

    def cargarCapaPostGIS( self, nombre, nombreCapa='', where='' ):
        """
            Funcin cargarCapaPostGIS
            -------------------------
            Agrega una capa de PostGIS al canvas. Soporta filtrado.

            nombre: Nombre de la tabla en PostGIS.
            nombreCapa: Alias de la capa para la aplicacin. Por defecto nulo.
            where: Filtro SQL. Por defecto nulo.
        """

        # Si no se obtuvo el parmetro nombreCapa asignarle el parm. nombre
        if nombreCapa == '': nombreCapa = str( nombre ).capitalize()

        uri = QgsDataSourceURI()
        # Host, puerto, bd, usuario, password
        uri.setConnection( SERVIDOR, PUERTO, BD, USUARIO, CONTRASENA )

        # Esquema, tabla, columna geometra y WHERE opcional
        uri.setDataSource( "public", nombre, "geom", where )

        layer = QgsVectorLayer( uri.uri(), nombreCapa, "postgres" )

        if layer.isValid():
            self.agregarCapa( layer )
        else:
            self.utils.mostrarMensaje( 'Advertencia',
                'No se ha podido establecer la conexin con la base de ' \
                + 'datos que especificaste.\n\nRevisa los parmetros de ' \
                + 'conexin.', QMessageBox.Warning )

    def agregarCapa( self, capa ):
        """ Agrega una nueva capa al canvas y especifica su simbologa """
        orden = 0 # Posicin de la nueva capa

        # Simbologa de la capa
        symbols = capa.renderer().symbols()
        symbol = symbols[ 0 ]

        if capa.name() == "Estacion" or capa.name() == "EtiquetasPortalIzq" \
            or capa.name() == "EtiquetasPortalDer": # Punto
            symbol.setPointSize( 1.3 )
            symbol.setLineWidth( 0.2 )
            symbol.setFillColor( QColor.fromRgb( 134,180,255 ) )
            symbol.setColor( QColor.fromRgb( 0,0,0 ) )
        elif capa.name() == "OrigenDestino": # Punto
            symbol.setPointSize( 1.7 )
            symbol.setLineWidth( 0.2 )
            symbol.setFillColor( QColor.fromRgb( 255,255,255 ) )
            symbol.setColor( QColor.fromRgb( 0,0,0 ) )
        elif capa.name() == "Parada": # Punto
            symbol.setPointSize( 1.3 )
            symbol.setLineWidth( 0.2 )
            symbol.setFillColor( QColor.fromRgb( 255,228,3 ) )
            symbol.setColor( QColor.fromRgb( 0,0,0 ) )
        elif capa.name() == "Segmento": # Linea
            symbol.setLineWidth( 0.8 )
            symbol.setColor( QColor.fromRgb( 161,176,162 ) )
        elif capa.name() == "Ruta": # Linea
            orden = 3 # Hay 3 capas por encima de esta
            symbol.setLineWidth( 1.7 )
            symbol.setColor( QColor.fromRgb( 255,0,0 ) )
        elif capa.name() == "Bogota":  #Poligono
            symbol.setLineWidth( 0 )
            symbol.setFillColor( QColor.fromRgb( 250,211,131 ) )
            symbol.setColor( QColor.fromRgb( 161,176,162 ) )

        # Agregar el layer al registro
        QgsMapLayerRegistry.instance().addMapLayer( capa )

        # Fijar extent del mapa al del primer layer cargado y fijar unidades
        if self.canvas.layerCount() == 0:
            self.canvas.setExtent( capa.extent() )
            self.canvas.setMapUnits( QGis.Meters )

        # Fijar el conjunto de capas (LayerSet) para el map canvas
        self.layers.insert( orden, QgsMapCanvasLayer( capa ) )
        self.canvas.setLayerSet( self.layers )

    def mostrarMapTip( self, p ):
        """ SLOT - Muestra en el mapa un Tooltip informativo """
        puntoQgs = QgsPoint( p.x(), p.y() )
        puntoTemp = self.mapToPixel.transform( puntoQgs.x(), puntoQgs.y() )
        puntoQt = QPoint( puntoTemp.x(), puntoTemp.y() )

        self.mapTip.showMapTip( self.capaEstacion, puntoQgs, puntoQt,
            self.canvas )

    def identificarEstacion( self, bBox ):
        """ SLOT - Identificar Estacin """
        self.grbIdentificar.setVisible( True )
        self.grbResultados.setVisible( False )
        self.grbAcercaDe.setVisible( False )
        self.grbMensajes.setVisible( False )

        provider = self.capaEstacion.dataProvider()
        posNombre = provider.fieldNameIndex( "nombreestacion" )
        posTipo = provider.fieldNameIndex( "tipoestacion" )

        provider.select( [ posNombre, posTipo ], bBox, False, False )
        feat = QgsFeature()
        aResultados = []

        # Llenar datos de la estacin
        while provider.nextFeature( feat ):
            attrs = feat.attributeMap()

            for ( k, attr ) in attrs.iteritems():
                if k == 1:
                    sRes1 = attr.toString()
                elif k == 2:
                    sRes2 = attr.toString()
            aResultados.append( [ sRes1, sRes2 ] )

        # Configurar la tabla de identificar
        titulos = [ 'NOMBRE DE LA ESTACIN', 'TIPO' ]
        model = ModeloTablas( aResultados, titulos, self ) # Modelo-Vista
        self.tblIdentificar.setModel( model )
        self.tblIdentificar.selectRow( 0 )
        self.tblIdentificar.setColumnWidth( 0,200 )
        self.tblIdentificar.setColumnWidth( 1,150 )
        self.tblIdentificar.resizeRowsToContents()

    def seleccionarEstacion( self, bBox ):
        """ SLOT - Seleccionar estacin """
        provider = self.capaEstacion.dataProvider()
        posNombre = provider.fieldNameIndex( "nombreestacion" )

        provider.select( [ posNombre ], bBox, False, False )
        feat = QgsFeature()
        aAttrs = []
        sEstacion = None

        while provider.nextFeature( feat ):
            attrs = feat.attributeMap()
            aAttrs.append( attrs[ 1 ].toString() )

        if len( aAttrs ) == 1:
            sEstacion = aAttrs[ 0 ]

        elif len( aAttrs ) > 1: # Seleccionar la estacin de un listado
            dialogo = DialogoSeleccion( self )
            aAttrs.sort()

            for i in range( len( aAttrs ) ):
                item = QListWidgetItem( dialogo.lista )
                item.setText( aAttrs[ i ] )

            dialogo.exec_()
            if dialogo.resultado:
                sEstacion = dialogo.resultado

        if sEstacion: # Si hubo resultado
            if not self.bSelOrigen:
                self.sOrigen = sEstacion
                self.cboOrigen.setCurrentIndex( self.cboOrigen.findText( self.sOrigen ) )
                self.bSelOrigen = True
                self.mostrarMensajeInferior( 2 )
            else:
                self.sDestino = sEstacion
                self.cboDestino.setCurrentIndex( self.cboDestino.findText( self.sDestino ) )
                self.bSelOrigen = False
                self.mostrarMensajeInferior( 3 )

    @pyqtSignature("")
    def nuevaRuta( self ):
        """ Click en el botn Crear Nueva Ruta """
        self.grbParametros.setVisible( True )
        self.grbAcercaDe.setVisible( True )
        self.grbMensajes.setVisible( False )
        self.grbNuevaRuta.setVisible( False )
        self.grbResultados.setVisible( False )
        self.grbIdentificar.setVisible( False )
        self.visualizarLeyenda( 1 ) # Mostrar la leyenda 1
        self.hacerVisiblesCapasRuta( False )
        self.btnSeleccionar.setEnabled( True )

        self.utils.mostrarLabels( self.utils.obtenerCapa( self.canvas,
            "EtiquetasPortalIzq" ), "nombreestacion" )
        self.utils.mostrarLabels( self.utils.obtenerCapa( self.canvas,
            "EtiquetasPortalDer" ), "nombreestacion" )

        self.canvas.zoomToFullExtent()

    def hacerVisiblesCapasRuta( self, accion ):
        """ Hace visibles o invisibles las capas de la ruta. accion:Booleano """
        for capa in self.layers:
            if capa.layer().name() == "Ruta" \
                or capa.layer().name() == "OrigenDestino" \
                or capa.layer().name() == "Parada":
                capa.setVisible( accion )

        self.canvas.setLayerSet(self.layers)

    def onChangecboOrigen( self, texto ):
        self.sOrigen = str( texto )

    def onChangecboDestino( self, texto ):
        self.sDestino = str( texto )

    def calcularRutas( self ):
        """ Lgica del negocio """
        self.bSelOrigen = False

        if self.sOrigen == self.sDestino:
            self.utils.mostrarMensaje( "Calcular rutas", "Las estaciones de" \
                + " Origen y Destino\ndeben ser diferentes.",
                QMessageBox.Warning )
            self.mostrarMensajeInferior( 1 )
        else:
            # Encontrar los servicios que pasan por las estaciones
            aServicios = self.conexionBD.consultarServicios( self.sOrigen,
                self.sDestino, self.cboDia.currentText(),
                self.cboHora.currentText(), self.cboMinutos.currentText() )

            if not aServicios:
                self.utils.mostrarMensaje( "Calcular rutas", "No hay rutas " \
                    + "que pasen el da definido por las estaciones " \
                    + "seleccionadas.", QMessageBox.Warning )
                self.mostrarMensajeInferior( 1 )
            else:
                self.progBar.show()
                self.grbResultados.setVisible( True )
                self.grbIdentificar.setVisible( False )
                self.grbMensajes.setVisible( False )
                self.grbAcercaDe.setVisible( False )
                self.grbParametros.setVisible( False )
                self.grbNuevaRuta.setVisible( True )
                self.btnSeleccionar.setEnabled( False )
                self.visualizarLeyenda( 2 ) # Mostrar la leyenda 2
                self.utils.ocultarLabels( self.utils.obtenerCapa( self.canvas,
                    "EtiquetasPortalIzq" ) )
                self.utils.ocultarLabels( self.utils.obtenerCapa( self.canvas,
                    "EtiquetasPortalDer" ) )

                # Desactivar botn de seleccin
                if self.btnSeleccionar.isChecked():
                    self.btnSeleccionar.setChecked( False )
                    self.canvas.unsetMapTool( self.toolSeleccionar )
                    self.canvas.setMapTool( self.toolAcercar )
                    self.btnAcercar.setChecked( True )

                aResultados = []
                self.idServicios = {}

                for i in range( len( aServicios ) ):
                    aResultado = []

                    # Diccionario de Ids
                    self.idServicios[ aServicios[i][1] ] = aServicios[i][0]

                    # Nombre de la Ruta
                    aResultado.append( aServicios[i][1] )

                    # Tiempo
                    aResultado.append( self.conexionBD.calcularTiempo( aServicios[i][0],
                        self.sOrigen, self.sDestino ) )

                    # Paradas
                    aResultado.append( self.conexionBD.calcularNoParadas( aServicios[i][0],
                        self.sOrigen, self.sDestino ) )

                    # Tipo de Servicio
                    aResultado.append( aServicios[i][2] )

                    # Origen
                    aResultado.append( self.sOrigen )

                    # Destino
                    aResultado.append( self.sDestino )

                    # Agregar resultado a arreglo
                    aResultados.append( aResultado )

                # Configurar la tabla de resultados
                titulos = [ 'RUTA', 'TIEMPO (min.)', 'PARADAS', 'SERVICIO',
                    'ORIGEN', 'DESTINO' ]
                aResultados = self.OrdenarResultados( aResultados )
                model = ModeloTablas( aResultados, titulos, self ) #Modelo-Vista
                self.tblResultados.setModel( model )
                self.tblResultados.setColumnWidth( 1,120 ) # Columna 'Tiempo'
                self.tblResultados.resizeRowsToContents()

                QObject.connect( self.tblResultados.selectionModel(),
                    SIGNAL( "selectionChanged(QItemSelection,QItemSelection)" ),
                    self.cambioSeleccion )

                self.progBar.setValue( 1 )
                self.tblResultados.selectRow( 0 )

    def OrdenarResultados( self, aRes ):
        """ Ordena los resultados por tiempo. La columna tiempo es la no. 2 """
        return sorted( aRes, key=operator.itemgetter( 1 ) )

    def cambioSeleccion( self, itemSelNuevo, itemSelViejo ):
        """ SLOT
            Carga y actualiza las capas de la ruta en el mapa al haber un cambio
            de seleccin en la tabla de resultados.

            itemSelNuevo: Conjunto de filas seleccionadas.
            itemSelViejo: Conjunto de filas que anteriormente estaban
                          seleccionadas.
        """
        fila = itemSelNuevo.indexes()[ 0 ].row()

        #Obtener el idServicio del diccionario
        sRuta = self.tblResultados.model().data( self.tblResultados.model().index( fila, 0 ),
            Qt.DisplayRole )
        idServicio = self.idServicios[ sRuta.toString() ]

        # Crear o actualizar vistas en PostGIS
        self.conexionBD.crearVistaEstaciones( self.sOrigen, self.sDestino )
        self.conexionBD.crearVistaParadas( idServicio, self.sOrigen, self.sDestino )
        self.conexionBD.crearVistaRuta( idServicio, self.sOrigen, self.sDestino )

        self.progBar.setValue( 2 )

        # Cargar o hacer visibles las capas de la Ruta
        if not self.bCapasRutaCreadas:
            self.cargarCapaPostGIS( "v_segmento", "Ruta" )
            self.progBar.setValue( 3 )
            self.cargarCapaPostGIS( "v_estacion", "OrigenDestino" )
            self.progBar.setValue( 4 )
            self.cargarCapaPostGIS( "v_parada", "Parada" )
            self.bCapasRutaCreadas = True
        else:
            self.hacerVisiblesCapasRuta( True )

        # Mostrar etiquetas
        self.utils.mostrarLabels( self.utils.obtenerCapa( self.canvas,
            "Parada" ), "nombreestacion" )
        self.utils.mostrarLabels( self.utils.obtenerCapa( self.canvas,
            "OrigenDestino" ), "nombreestacion" )
        self.progBar.setValue( 5 )
        self.canvas.refresh()

        self.progBar.hide()

    def cargarCombosOrigenDestino( self ):
        """ Inicializa los combos Origen y Destino """
        aNombres = self.conexionBD.consultarNombresEstaciones()
        aNombres.sort()

        if aNombres:
            self.cboOrigen.addItems( aNombres )
            self.cboDestino.addItems( aNombres )
            self.cboDestino.setCurrentIndex( 4 ) # Inicializar

    def mostrarMensajeInferior( self, numero ):
        """ Actualiza el mensaje inferior de la aplicacin """
        if numero == 1:
            self.lblMensaje.setText( "1. Selecciona la estacin de " \
                + "<b>origen</b> en el mapa." )
        elif numero == 2:
            self.lblMensaje.setText( "2. Ahora selecciona la estacin de " \
                + "<b>destino</b> en el mapa." )
        else:
            self.lblMensaje.setText( "3. Selecciona el da y la hora del " \
                + "viaje en la parte superior de la ventana<br />y haz click " \
                + "en el botn <b>Calcular Rutas.</b>" )

    def cargarLeyenda( self ):
        """ Carga la leyenda del mapa """
        self.lblLeyenda1 = QLabel( self.frameMapa )
        self.lblLeyenda1.setGeometry( QRect( 560,317,54,47 ) )
        self.lblLeyenda1.setFrameShape( QFrame.Box )
        self.lblLeyenda1.setFrameShadow( QFrame.Raised )
        self.lblLeyenda1.setPixmap( QPixmap( ":/otros/imagenes/leyenda1.png" ) )
        self.lblLeyenda1.setMargin( 1 )
        self.lblLeyenda1.setObjectName( "lblLeyenda1" )

        self.lblLeyenda2 = QLabel( self.frameMapa )
        self.lblLeyenda2.setGeometry( QRect( 560,300,54,64 ) )
        self.lblLeyenda2.setFrameShape( QFrame.Box )
        self.lblLeyenda2.setFrameShadow( QFrame.Raised )
        self.lblLeyenda2.setPixmap( QPixmap( ":/otros/imagenes/leyenda2.png" ) )
        self.lblLeyenda2.setMargin( 1 )
        self.lblLeyenda2.setObjectName( "lblLeyenda2" )
        self.lblLeyenda2.setVisible( False )

    def cargarNorte( self ):
        """ Carga la imagen que indica el Norte """
        self.lblNorte = QLabel( self.frameMapa )
        self.lblNorte.setGeometry( QRect( 575,22,40,40 ) )
        self.lblNorte.setPixmap( QPixmap( ":/otros/imagenes/norte.png" ) )
        self.lblNorte.setObjectName( "lblNorte" )

    def cargarProgressBar( self ):
        """ Cargar la barra de progreso """
        self.progBar = QProgressBar( self.frameMapa )
        self.progBar.setGeometry( QRect( 25,25,141,23 ) )
        self.progBar.setMaximum( 5 )
        self.progBar.setValue( 0 )
        self.progBar.setObjectName( "progBar" )
        self.progBar.hide()

    def visualizarLeyenda( self, numero ):
        """ Visualiza la leyenda del mapa segn el nmero """
        if numero == 1: # Leyenda inicial
            self.lblLeyenda1.setVisible( True )
            self.lblLeyenda2.setVisible( False )
        else: # Leyenda completa
            self.lblLeyenda2.setVisible( True )
            self.lblLeyenda1.setVisible( False )

    def closeEvent( self, event ):
        """ Previene cerrar por accidente la aplicacin """
        res = self.utils.mostrarMensajeConfirmacion( "TrasMiSIG",
            "Realmente quieres salir?" )
        if res == "si":
            event.accept()
        elif res == "no":
            event.ignore()


class MapTool( QgsMapTool ):
    """ Implementa la clase MapTool para identificar o seleccionar estaciones """
    def __init__( self, canvas, tipo ):
        QgsMapTool.__init__( self, canvas )
        self.tipo = tipo # Identificar, Seleccionar

    def canvasReleaseEvent( self, event ):
        p = QPoint( event.pos().x(), event.pos().y() )
        tolerancia = 4
        pmin = self.toMapCoordinates( QPoint( p.x() - tolerancia,
            p.y() - tolerancia ) )
        pmax = self.toMapCoordinates( QPoint(p.x() + tolerancia,
            p.y() + tolerancia ) )
        bBox = QgsRectangle( pmin.x(), pmin.y(), pmax.x(), pmax.y() )

        if self.tipo == "Identificar":
            self.emit( SIGNAL( "identificarEstacion" ), bBox )
        elif self.tipo == "Seleccionar":
            self.emit( SIGNAL( "seleccionarEstacion" ), bBox )


class Utilidades():
    """ Provee funciones comunes """
    def __init__( self, parent=None ):
        self.parent = parent
        self.ultimaRutaExportarMapaAImagen = "."

    def mostrarMensaje( self, titulo, desc, icono=None ):
        """ Muestra mensajes al usuario """
        msg = QMessageBox( self.parent )
        msg.setText( desc )
        msg.setWindowTitle( titulo )
        if icono:
            msg.setIcon( icono )
        msg.addButton( 'Cerrar', QMessageBox.RejectRole )
        msg.exec_()

    def mostrarMensajeConfirmacion( self, titulo, desc ):
        """ Muestra un mensaje de confirmacin al usuario """
        msg = QMessageBox( self.parent )
        msg.setText( desc )
        msg.setWindowTitle( titulo )
        msg.setIcon( QMessageBox.Question )

        botonSi = msg.addButton( 'Si', QMessageBox.YesRole )
        botonNo = msg.addButton( 'No', QMessageBox.NoRole )
        msg.exec_()

        if msg.clickedButton() == botonSi:
            return "si"
        else:
            return "no"

    def mostrarLabels( self, capa, campo ):
        """
            Funcin mostrarLabels
            ---------------------
            Muestra etiquetas para la capa especificada.

            capa: Objeto capa.
            campo: Nombre del campo base para las etiquetas.
        """
        label = capa.label()
        labelAtt = label.layerAttributes()

        # Definir campo para etiquetar
        idCampo = capa.dataProvider().fieldNameIndex( campo )
        label.setLabelField( QgsLabel.Text, idCampo )

        labelAtt.setText( '' ) # Texto por defecto para la etiqueta

        if capa.name() == "Parada":
            labelAtt.setColor( Qt.black )
            labelAtt.setSize( 6.5, QgsLabelAttributes.PointUnits )
            labelAtt.setAlignment( Qt.AlignLeft | Qt.AlignTop )
            labelAtt.setOffset( 2, 4, QgsLabelAttributes.PointUnits )
            label.setMaxScale( 125000 )
            label.setScaleBasedVisibility( True )
        else:
            labelAtt.setColor( Qt.blue )
            labelAtt.setBufferEnabled( True )
            labelAtt.setBufferColor( Qt.white )
            labelAtt.setBufferSize( 1, QgsLabelAttributes.PointUnits )
            labelAtt.setSize( 7, QgsLabelAttributes.PointUnits )
            labelAtt.setBold( True )
            labelAtt.setAlignment( Qt.AlignRight | Qt.AlignBottom )
            labelAtt.setOffset( -4, 2, QgsLabelAttributes.PointUnits )

        if capa.name() == "EtiquetasPortalDer":
            labelAtt.setAlignment( Qt.AlignLeft | Qt.AlignBottom )
            labelAtt.setOffset( 4, 2, QgsLabelAttributes.PointUnits )

        capa.enableLabels( True )

    def ocultarLabels( self, capa ):
        """ Oculta las etiquetas de la capa especificada """
        self.mostrarLabels( capa, "###" )  # Por mejorar! Aqu se est pasando un
                                         # nombre de un campo que no existe.

    def obtenerCapa( self, canvas, nombre ):
        """ Retorna un objeto capa a partir de su nombre """
        for i in range( canvas.layerCount() ):
            if canvas.layer( i ).name() == nombre:
                return canvas.layer( i )

class DialogoSeleccion( QDialog ):
    """ Provee un dilogo para seleccionar una estacin entre varias opciones """
    def __init__( self, parent=None ):
        QDialog.__init__( self, parent )

        self.setWindowTitle( 'Seleccionar estacin' )

        lblMensaje = QLabel( 'Selecciona una de las estaciones:' )
        lstWidget = QListWidget()

        font = QFont()
        font.setFamily( "Arial" )
        font.setPointSize( 10 )
        lstWidget.setFont( font )
        lstWidget.setHorizontalScrollBarPolicy( Qt.ScrollBarAlwaysOff )
        lstWidget.setAlternatingRowColors( True )
        lstWidget.setSelectionMode( QAbstractItemView.SingleSelection )

        lstWidget.setViewMode( QListView.ListMode )
        lstWidget.setUniformItemSizes( True )
        lstWidget.setObjectName( "lstWidget" )

        buttonBox = QDialogButtonBox()
        self.btnAceptar = buttonBox.addButton( 'Aceptar',
            QDialogButtonBox.AcceptRole )
        buttonBox.addButton( 'Cancelar', QDialogButtonBox.RejectRole )

        QObject.connect( buttonBox, SIGNAL( "accepted()" ), self.aceptar )
        QObject.connect( buttonBox, SIGNAL( "rejected()" ), self.cancelar )

        grid = QGridLayout()
        grid.setSpacing( 10 )
        grid.addWidget( lblMensaje, 1, 0, 1, 2, Qt.AlignCenter )
        grid.addWidget( lstWidget, 2, 0, 1, 2, Qt.AlignCenter )
        grid.setRowStretch( 3, 1 )
        grid.addWidget( buttonBox, 4, 0, 1, 2, Qt.AlignCenter )

        self.setLayout( grid )
        self.setFixedSize( 260, 170 )
        self.lista = lstWidget

    def aceptar( self ):
        self.resultado = self.lista.currentItem().text()
        self.close()

    def cancelar( self ):
        self.resultado = None
        self.close()


class ModeloTablas( QAbstractTableModel ):
    """
        Clase ModeloTablas
        ------------------
        Provee el modelo de almacenamiento para las tablas.
        Implementa el Modelo Vista.
    """
    def __init__( self, arreglo, encabezados, parent=None, *args ):
        QAbstractTableModel.__init__( self, parent, *args )
        self.datos = arreglo
        self.headerdata = encabezados
        self.numFilas = len( self.datos )

    def rowCount( self, parent ):
        """ Retorna el nmero de filas en la tabla """
        return len( self.datos )

    def columnCount( self, parent ):
        """ Retorna el nmero de columnas en la tabla """
        if len( self.datos ) > 0:
            return len( self.datos[ 0 ] )
        else:
            return 0

    def data(self, index, role):
        """ Retorna los datos de la tabla """
        if role == Qt.DisplayRole:
            return QVariant( self.datos[ index.row() ][ index.column() ] )
        elif role ==  Qt.TextAlignmentRole: # Posicin del texto en la celda
            return QVariant( Qt.AlignCenter )

        return QVariant()

    def headerData( self, section, orientation, role ):
        """ Retorna los encabezados de la tabla """
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return QVariant( self.headerdata[ section ] )
        if orientation == Qt.Vertical and role == Qt.DisplayRole:
            # Numera los encabezados verticales
            return QVariant( QString( "%1" ).arg( section + 1 ) )
        return QVariant()